:INPUT,FORWARD,OUTPUT
-m srh --srh-next-hdr 17;=;OK
-m srh --srh-hdr-len-eq 8;=;OK
-m srh --srh-hdr-len-gt 8;=;OK
-m srh --srh-hdr-len-lt 8;=;OK
-m srh --srh-segs-left-eq 1;=;OK
-m srh --srh-segs-left-gt 1;=;OK
-m srh --srh-segs-left-lt 1;=;OK
-m srh --srh-last-entry-eq 4;=;OK
-m srh --srh-last-entry-gt 4;=;OK
-m srh --srh-last-entry-lt 4;=;OK
-m srh --srh-tag 0;=;OK
-m srh ! --srh-next-hdr 17;=;OK
-m srh ! --srh-hdr-len-eq 8;=;OK
-m srh ! --srh-hdr-len-gt 8;=;OK
-m srh ! --srh-hdr-len-lt 8;=;OK
-m srh ! --srh-segs-left-eq 1;=;OK
-m srh ! --srh-segs-left-gt 1;=;OK
-m srh ! --srh-segs-left-lt 1;=;OK
-m srh ! --srh-last-entry-eq 4;=;OK
-m srh ! --srh-last-entry-gt 4;=;OK
-m srh ! --srh-last-entry-lt 4;=;OK
-m srh ! --srh-tag 0;=;OK
-m srh --srh-next-hdr 17 --srh-segs-left-eq 1 --srh-last-entry-eq 4 --srh-tag 0;=;OK
-m srh ! --srh-next-hdr 17 ! --srh-segs-left-eq 0 --srh-tag 0;=;OK
-m srh --srh-psid A::/64 --srh-nsid B:: --srh-lsid C::/0;;OK
-m srh ! --srh-psid A::/64 ! --srh-nsid B:: ! --srh-lsid C::/0;;OK
-m srh;=;OK
